# file_outliers.py
# Python 3.7.6  NumPy 1.18.1

import numpy as np

def line_count(fn):
  ct = 0
  fin = open(fn, "r")
  for line in fin:
    ct += 1
  fin.close()
  return ct

def show_file(fn, start, end, indices=False,
 strip_nl=False):
  fin = open(fn, "r")

  # advance to start line
  ln = 1
  while ln < start:
    fin.readline()
    ln += 1

  # show remaining lines
  while ln <= end:
    line = fin.readline()
    if line == "": break  # EOF
    if strip_nl == True:
      line = line.strip()
    if indices == True:
      print("[%3d]  " % ln, end="")
    print(line)
    ln += 1
  fin.close()

def delete_lines(src, dest, omit_lines):
  fin = open(src, "r"); fout = open(dest, "w")
  line_num = 1
  for line in fin:
    if line_num in omit_lines:
      line_num += 1 
    else:
      fout.write(line)  # has embedded nl
      line_num += 1
  fout.close(); fin.close()

def show_numeric_outliers(fn, col, z_max, delim):
  # need 3 passes so read into memory
  ct = line_count(fn)
  data = np.empty(ct, dtype=np.object)

  fin = open(fn, "r")
  i = 0
  for line in fin:
    data[i] = line
    i += 1
  fin.close()

  # compute mean, sd of a 1-based col
  sum = 0.0
  for i in range(len(data)):
    line = data[i]
    tokens = line.split(delim)
    sum += np.float32(tokens[col-1])
  mean = sum / len(data)

  ss = 0.0
  for i in range(len(data)):
    line = data[i]
    tokens = line.split(delim)
    x = np.float32(tokens[col-1])
    ss += (x - mean) * (x - mean)
  sd = np.sqrt(ss / len(data))  # population sd
  print("mean = %0.2f" % mean)
  print("sd   = %0.2f" % sd)

  # display outliers
  for i in range(len(data)):
    line = data[i]
    tokens = line.split(delim)
    x = np.float32(tokens[col-1])
    z = (x - mean) /sd
    if z <= -z_max or z >= z_max:
      print("[%3d] x = %0.4f z = %0.2f" % \
        ((i+1), x, z))

def show_cat_outliers(fn, col, ct_min, delim):
  # need 2 passes so read into memory
  ct = line_count(fn)
  data = np.empty(ct, dtype=np.object)

  fin = open(fn, "r")
  i = 0
  for line in fin:
    data[i] = line
    i += 1
  fin.close()

  # construct dictionary for column
  d = dict()
  for i in range(len(data)):
    line = data[i]
    tokens = line.split(delim)
    sv = tokens[col-1]  # string value
    if sv not in d:
      d[sv] = 1
    else:
      d[sv] += 1

  print("\nValues \t Counts")
  for (sv,  ct) in d.items():
    print("%s \t %d" % (sv, ct))

  print("\nRare values:")
  for i in range(len(data)):
    line = data[i]
    tokens = line.split(delim)
    sv = tokens[col-1]  # string value
    ct = d[sv]  # get count  
    if ct <= ct_min:
      print("[%3d] cat value = %s \
        count = %d" % ((i+1), sv, ct)) 

def update_line(src, dest, line_num, col_num,
  new_val, delim):
  # line_num and col_num are 1-based
  fin = open(src, "r")
  fout = open(dest, "w")
  ln = 1

  while ln < line_num:
    line = fin.readline()  # has embedded nl
    fout.write(line)
    ln += 1

  line_to_update = fin.readline()  # has trailing nl
  tokens = line_to_update.split(delim)
  tokens[col_num-1] = new_val  # tokens are 0-based

  s = ""
  for j in range(len(tokens)):  # including trailing nl
    if j == len(tokens)-1:  # at the newline
      s += tokens[j]
    else:
      s += tokens[j] + delim  # interior column
  fout.write(s)  # must have embedded nl

  for line in fin:
    fout.write(line)  # remaining lines

  fout.close(); fin.close()

def main():
  # 1. display source file
  print("\nSource file: ")
  fn = ".\\people_no_missing.txt"
  show_file(fn, 1, 999, indices=True, strip_nl=True)

  # 2. numeric outliers
  print("\nIdentifying outliers in Age column:")
  fn = ".\\people_no_missing.txt"
  show_numeric_outliers(fn, 2, 2.0, "\t")  # age

  print("\nModifying line [9] to age = 33")
  src = ".\\people_no_missing.txt"
  dest = ".\\people_no_missing_update1.txt"
  update_line(src, dest, 9, 2, "33", "\t")  # line 9, col 2

  # 3. categorical outliers
  print("\nExamining Region column:")
  fn = ".\\people_no_missing_update1.txt" 
  show_cat_outliers(fn, 3, ct_min=1, delim="\t")

  print("\nUpdating line [6], deleting line [4]")
  src = ".\\people_no_missing_update1.txt"
  dest = ".\\people_no_missing_update2.txt"
  update_line(src, dest, 6, 3, "central", "\t")

  src = ".\\people_no_missing_update2.txt"
  dest = ".\\people_clean.txt"
  delete_lines(src, dest, [4])

  print("\nCleaned data: ")
  fn = ".\\people_clean.txt"
  show_file(fn, 1, 999, indices=True, strip_nl=True)

if __name__ == "__main__":
  main()

